#  Copyright 2020 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
import sys


class LoggingHelper:

    @classmethod
    def configure_for_cli(cls):
        cls.configure_for_test()

    @classmethod
    def configure_for_test(cls):
        logging.getLogger('urllib3').setLevel(logging.WARNING)
        logging.getLogger('botocore').setLevel(logging.WARNING)
        logging.getLogger('pyathena').setLevel(logging.WARNING)
        logging.getLogger('faker').setLevel(logging.ERROR)
        logging.getLogger('snowflake').setLevel(logging.WARNING)
        logging.basicConfig(
            level=logging.DEBUG,
            # https://docs.python.org/3/library/logging.html#logrecord-attributes
            # %(name)s
            format="  | %(message)s",
            handlers=[logging.StreamHandler(sys.stdout)]
        )
