#  Copyright 2020 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from typing import List, Optional


class Missing:

    FORMATS = {
        # more regexes: https://regexr.com/ https://digitalfortress.tech/tricks/top-15-commonly-used-regex/ http://regexlib.com/
        'empty': r'^$',
        'whitespace': r'^\s*$'
    }

    def __init__(self):
        self.values: Optional[List[str]] = None
        self.format: Optional[str] = None
        self.regex: Optional[str] = None
